/*
 * @(#) $Id: fourier-compose.c 334 2006-01-02 23:14:46Z yeti $
 * This is a plugin for the GIMP.
 *
 * Copyright (C) 2005 Yeti (David Necas) <yeti@physics.muni.cz>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*
 * This plug-in generates tileable textures using Fourier synthesis
 *
 * Please send patches, suggestions, and other invectives to me:
 * <yeti@physics.muni.cz>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#else
/* To include libgimp/stdplugins-intl.h */
#  define GETTEXT_PACKAGE "gimp20"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <fftw3.h>
#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include "libgimp/stdplugins-intl.h"

#define PROCEDURE_NAME   "fourier_compose"
#define DATA_KEY_VALS    "plug_in_fourier_compose"

enum {
    BITS_PER_SAMPLE   = 8,
    CHANNEL_MAX_VALUE = (1U << BITS_PER_SAMPLE) - 1,

    PREVIEW_SIZE = 128,
    SCALE_WIDTH = 201,

    /* FIXME: Is it still impossible to store true unlimited strings as
     * parameters? */
    GRADIENT_NAME_SIZE = 256,
    GRADIENT_SAMPLE_SIZE = 8192,

    /* How often progressbar is updated */
    PROGRESS_STEP = 0x2000
};

enum {
    RESPONSE_REVERT = 1,
    RESPONSE_RESET
};

enum {
    RECOMPUTE_ALL   = 1 << 0,
    RECOMPUTE_COLOR = 1 << 1
};

typedef enum {
    COLORING_METHOD_GRAY     = 0,
    COLORING_METHOD_GRADIENT = 1
} ColoringMethod;

typedef struct {
    gdouble        freq_min;
    gdouble        freq_max;
    gboolean       gauss_enable;
    gdouble        gauss_0;
    gdouble        gauss_sigma;
    gboolean       power_enable;
    gdouble        power;
    gboolean       dir_enable;
    gint           dir_n;
    gdouble        dir_strength;
    gdouble        dir_rot;
    gint           seed;
    ColoringMethod cm;
    gchar          grad[GRADIENT_NAME_SIZE];
    gboolean       random_seed;
} PluginValues;

typedef struct {
    GtkWidget *seed;
    GtkWidget *cm_gray;
    GtkWidget *cm_grad;
    GtkWidget *grad;
    GtkObject *freq_min;
    GtkObject *freq_max;
    GtkWidget *gauss_enable;
    GtkObject *gauss_0;
    GtkObject *gauss_sigma;
    GtkWidget *power_enable;
    GtkObject *power;
    GtkWidget *dir_enable;
    GtkObject *dir_n;
    GtkObject *dir_strength;
    GtkObject *dir_rot;
} PluginControls;

static void     query                  (void);
static void     run                    (const gchar *name,
                                        gint nparams,
                                        const GimpParam *param,
                                        gint *nreturn_vals,
                                        GimpParam **return_vals);
static gboolean plugin_dialog          (void);
static void     gauss_enable_toggled   (GtkWidget *check);
static void     power_enable_toggled   (GtkWidget *check);
static void     dir_enable_toggled     (GtkWidget *check);
static void     coloring_method_changed(GtkWidget *widget,
                                        gpointer data);
static void     gradient_changed       (const gchar *name,
                                        gint width,
                                        const gdouble *grad_data,
                                        gboolean dialog_closing,
                                        gpointer user_data);
static void     refresh_controls       (const PluginValues *new_pvals);
static void     execute_all            (GimpDrawable *drawable);
static void     execute_color          (GimpDrawable *drawable);
static void     execute                (GimpDrawable *drawable,
                                        guint recompute);
static void     commit                 (GimpDrawable *drawable,
                                        gint chmax);
static void     commit_one_tile        (GimpPixelRgn *rgn,
                                        gint chmax,
                                        gint gradwidth);
static void     render_gradient        (void);

static GimpPlugInInfo PLUG_IN_INFO = {
    NULL,  /* init_proc  */
    NULL,  /* quit_proc  */
    query, /* query_proc */
    run,   /* run_proc   */
};

static const PluginValues pvals_defaults = {
    0.0,                      /* freq_min     */
    1.0,                      /* freq_max     */
    FALSE,                    /* gauss_enable */
    0.0,                      /* gauss_0      */
    0.03,                     /* gauss_sigma  */
    FALSE,                    /* power_enable */
    1.6,                      /* power        */
    FALSE,                    /* dir_enable   */
    5,                        /* dir_n        */
    1.0,                      /* dir_strength */
    0.0,                      /* dir_rot      */
    11,                       /* seed         */
    COLORING_METHOD_GRADIENT, /* cm           */
    "",                       /* grad         */
    0,                        /* random_seed  */
};

static PluginControls pctrl;
static PluginValues pvals; /* = pvals_defaults; */
static PluginValues pvals_old; /* for revert */

static GRand *rng = NULL;
static GtkWidget *preview_image = NULL;
static GdkPixbuf *preview = NULL;
static gboolean preview_mode;
static gboolean is_rgb;
static GimpDrawable *drawable;
static gdouble *gradient = NULL;
static gint gradient_p_width = -1;
static gboolean in_param_reset = TRUE;

static gint    ix1, iy1, ix2, iy2;     /* Selected image size. */
static gint    bpp, has_alpha, alpha;
static glong   progress;
static gdouble max_progress;  /* saves typecasting to double */
static fftw_plan fft_plan;
static fftw_complex *fft_data = NULL;


MAIN()

static void
query(void)
{
    static GimpParamDef args[] = {
        { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
        { GIMP_PDB_IMAGE, "image", "Input image (unused)" },
        { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
        { GIMP_PDB_INT32, "seed", "Random seed" },
        { GIMP_PDB_FLOAT, "freq_min", "Minimum frequency" },
        { GIMP_PDB_FLOAT, "freq_max", "Maximum frequency" },
        { GIMP_PDB_INT32, "gauss_enable", "Whether Gaussian multiplier is enabled" },
        { GIMP_PDB_FLOAT, "gauss_0", "Gaussian multiplier center" },
        { GIMP_PDB_FLOAT, "gauss_sigma", "Gaussian multiplier sigma" },
        { GIMP_PDB_INT32, "power_enable", "Whether power multiplier is enabled" },
        { GIMP_PDB_FLOAT, "power", "Inverse power multiplier power" },
        { GIMP_PDB_INT32, "dir_enable", "Whether directional multiplier is enabled" },
        { GIMP_PDB_INT32, "dir_n", "Number of directions in directional multiplier" },
        { GIMP_PDB_FLOAT, "dir_strength", "Strength of directional multiplier" },
        { GIMP_PDB_FLOAT, "dir_rot", "Rotation of directional multiplier" },
        { GIMP_PDB_INT32, "cm", "Coloring method: { GRAYSCALE (0), GRADIENT (1) }" },
        { GIMP_PDB_STRING, "grad", "Gradient used for coloring (when cm is GRADIENT)" },
    };

    gimp_install_procedure
        (DATA_KEY_VALS,
         "Generates a texture using Fourier synthesis",
         "This plugin generates a tileable texture using Fourier synthesis "
         "with various options.",
         "Yeti",
         "Yeti",
         "Dec 2005",
         N_("<Image>/Filters/Render/_Fourier Compose..."),
         "RGB*, GRAY*",
         GIMP_PLUGIN,
         G_N_ELEMENTS(args), 0, args, NULL);
}

static void
run(const gchar *name,
    gint nparams,
    const GimpParam *param,
    gint *nreturn_vals,
    GimpParam **return_vals)
{
    static GimpParam values[1];
    GimpRunMode run_mode;
    GimpPDBStatusType status = GIMP_PDB_SUCCESS;

    run_mode = param[0].data.d_int32;

    bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
#endif
    textdomain(GETTEXT_PACKAGE);

    *nreturn_vals = 1;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    drawable = gimp_drawable_get(param[2].data.d_drawable);
    is_rgb = gimp_drawable_is_rgb(drawable->drawable_id);
    has_alpha = gimp_drawable_has_alpha(drawable->drawable_id);

    pvals = pvals_defaults;
    rng = g_rand_new();
    switch (run_mode) {
        case GIMP_RUN_INTERACTIVE:
        gimp_get_data(DATA_KEY_VALS, &pvals);

        pvals_old = pvals;

        if (!plugin_dialog()) {
            gimp_drawable_detach(drawable);
            return;
        }
        break;

        case GIMP_RUN_NONINTERACTIVE:
        if (nparams != 17)
            status = GIMP_PDB_CALLING_ERROR;
        else {
            pvals.seed = (gint)param[3].data.d_int32;
            pvals.freq_min = (gdouble)param[4].data.d_float;
            pvals.freq_max = (gdouble)param[5].data.d_float;
            pvals.gauss_enable = (gboolean)param[6].data.d_int32;
            pvals.gauss_0 = (gdouble)param[7].data.d_float;
            pvals.gauss_sigma = (gdouble)param[8].data.d_float;
            pvals.power_enable = (gboolean)param[8].data.d_int32;
            pvals.power = (gdouble)param[10].data.d_float;
            pvals.dir_enable = (gboolean)param[11].data.d_int32;
            pvals.dir_n = (gboolean)param[12].data.d_int32;
            pvals.dir_strength = (gdouble)param[13].data.d_float;
            pvals.dir_rot = (gdouble)param[14].data.d_float;
            pvals.cm = (ColoringMethod)param[15].data.d_int32;
            g_strlcpy(pvals.grad, (gchar*)param[14].data.d_string,
                      sizeof(pvals.grad));
        }
        break;

        case GIMP_RUN_WITH_LAST_VALS:
        gimp_get_data(DATA_KEY_VALS, &pvals);
        break;

        default:
        break;
    }

    if (status == GIMP_PDB_SUCCESS) {
        if (is_rgb || gimp_drawable_is_gray(drawable->drawable_id)) {
            gimp_progress_init(_("Fourier Compose..."));

            preview_mode = FALSE;
            in_param_reset = FALSE;
            execute(drawable, RECOMPUTE_ALL);

            if (run_mode != GIMP_RUN_NONINTERACTIVE)
                gimp_displays_flush();

            if (run_mode == GIMP_RUN_INTERACTIVE
                || run_mode == GIMP_RUN_WITH_LAST_VALS) {
                g_strlcpy(pvals.grad, gimp_context_get_gradient(),
                          sizeof(pvals.grad));
                gimp_set_data(DATA_KEY_VALS, &pvals, sizeof(PluginValues));
            }
        }
        else
            status = GIMP_PDB_EXECUTION_ERROR;
    }

    values[0].data.d_status = status;
    gimp_drawable_detach(drawable);
}

/* remove some source code redundancies {{{ */
static inline void
yeti_table_attach_to_row(GtkWidget *table,
                         GtkWidget *child,
                         gint row)
{
    gtk_table_attach(GTK_TABLE(table), child, 0, GTK_TABLE(table)->ncols,
                     row, row+1, GTK_EXPAND|GTK_FILL, 0, 0, 0);
}

static inline void
yeti_table_attach_single_cell(GtkWidget *table,
                              GtkWidget *child,
                              gint column,
                              gint row)
{
    gtk_table_attach(GTK_TABLE(table), child, column, column+1, row, row+1,
                     GTK_EXPAND|GTK_FILL, 0, 0, 0);
}

static inline GtkWidget*
yeti_preview_new(GtkWidget *hbox,
                 GimpDrawable *drawable)
{
    GtkWidget *frame, *vbox, *image;
    GdkPixbuf *pixbuf;

    vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);

    pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB,
                            FALSE,
                            BITS_PER_SAMPLE,
                            PREVIEW_SIZE,
                            PREVIEW_SIZE);
    gdk_pixbuf_fill(pixbuf, 0);

    image = gtk_image_new_from_pixbuf(pixbuf);
    g_object_unref(pixbuf);
    gtk_container_add(GTK_CONTAINER(frame), image);

    return image;
}

static inline GtkObject*
yeti_scale_entry_new_double(const gchar *name,
                            GtkWidget *table,
                            GtkSizeGroup *sgroup,
                            gint row,
                            gdouble *value,
                            gdouble min,
                            gdouble max,
                            gdouble step,
                            gdouble page,
                            gint digits)
{
    GtkObject *adj;

    *value = CLAMP(*value, min, max);
    adj = gimp_scale_entry_new(GTK_TABLE(table), 0, row, name, SCALE_WIDTH, 0,
                               *value, min, max, step, page, digits,
                               TRUE, 0, 0, NULL, NULL);
    g_signal_connect(adj, "value-changed",
                     G_CALLBACK(gimp_double_adjustment_update), value);
    if (sgroup)
        gtk_size_group_add_widget(sgroup, GIMP_SCALE_ENTRY_SPINBUTTON(adj));

    return adj;
}

static inline GtkObject*
yeti_scale_entry_new_int(const gchar *name,
                         GtkWidget *table,
                         GtkSizeGroup *sgroup,
                         gint row,
                         gint *value,
                         gint min,
                         gint max)
{
    GtkObject *adj;

    adj = gimp_scale_entry_new(GTK_TABLE(table), 0, row, name, SCALE_WIDTH, 0,
                               *value, min, max, 1, 10, 0,
                               TRUE, 0, 0, NULL, NULL);
    g_signal_connect(G_OBJECT(adj), "value-changed",
                     G_CALLBACK(gimp_int_adjustment_update),
                     value);
    if (sgroup)
        gtk_size_group_add_widget(sgroup, GIMP_SCALE_ENTRY_SPINBUTTON(adj));

    return adj;
}

static inline void
yeti_coloring_method_attach(GtkWidget *table,
                            gint *row,
                            const gchar *gradlabel,
                            GtkWidget **grad,
                            GtkWidget **cm_gray,
                            GtkWidget **cm_grad,
                            ColoringMethod *cm)
{
    GtkRadioButton *group;
    GtkWidget *align;

    if (!is_rgb)
        *cm = COLORING_METHOD_GRAY;

    *grad = gimp_gradient_select_widget_new(gradlabel,
                                            gimp_context_get_gradient(),
                                            gradient_changed,
                                            NULL);
    align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
    gtk_container_add(GTK_CONTAINER(align), *grad);
    yeti_table_attach_single_cell(table, align, 1, *row+1);
    gtk_widget_set_sensitive(*grad, *cm == COLORING_METHOD_GRADIENT);

    *cm_gray = gtk_radio_button_new_with_mnemonic(NULL, _("Gra_yscale"));
    group = GTK_RADIO_BUTTON(*cm_gray);
    g_signal_connect(*cm_gray, "toggled",
                     G_CALLBACK(coloring_method_changed),
                     GUINT_TO_POINTER(COLORING_METHOD_GRAY));
    yeti_table_attach_single_cell(table, *cm_gray, 0, *row);

    *cm_grad = gtk_radio_button_new_with_mnemonic_from_widget(group,
                                                              _("_Gradient"));
    if (*cm == COLORING_METHOD_GRADIENT)
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(*cm_grad), TRUE);
    g_signal_connect(*cm_grad, "toggled",
                     G_CALLBACK(coloring_method_changed),
                     GUINT_TO_POINTER(COLORING_METHOD_GRADIENT));
    yeti_table_attach_single_cell(table, *cm_grad, 0, *row+1);
    if (!is_rgb)
        gtk_widget_set_sensitive(*cm_grad, FALSE);

    *row += 2;
}
/* }}} */

static gboolean
plugin_dialog(void)
{
    GtkWidget *dlg, *main_vbox, *hbox, *table, *check, *align;
    GtkSizeGroup *sgroup;
    gint row, run;
    gboolean ok;

    preview_mode = TRUE;
    in_param_reset = TRUE;
    gimp_ui_init(PROCEDURE_NAME, TRUE);

    dlg = gimp_dialog_new(_("Fourier Compose"), PROCEDURE_NAME,
                          NULL, 0,
                          gimp_standard_help_func,
                          "filters/fourier_compose.html",

                          _("Re_vert"), RESPONSE_REVERT,
                          GIMP_STOCK_RESET, RESPONSE_RESET,
                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                          GTK_STOCK_OK, GTK_RESPONSE_OK,
                          NULL);

    sgroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

    main_vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 6);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), main_vbox,
                       TRUE, TRUE, 0);

    /* top part with preview and common options */
    hbox = gtk_hbox_new(FALSE, 8);
    gtk_box_pack_start(GTK_BOX(main_vbox), hbox, TRUE, TRUE, 0);

    preview_image = yeti_preview_new(hbox, drawable);
    preview = gtk_image_get_pixbuf(GTK_IMAGE(preview_image));

    table = gtk_table_new(3, 4, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    row = 0;

    align = gtk_alignment_new(0.0, 0.0, 1.0, 0.0);
    gtk_container_add(GTK_CONTAINER(align), table);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);

    yeti_coloring_method_attach(table, &row, _("Fourier Compose Gradient"),
                                &pctrl.grad, &pctrl.cm_gray, &pctrl.cm_grad,
                                &pvals.cm);

    table = gtk_table_new(14, 3, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_box_pack_start(GTK_BOX(main_vbox), table, FALSE, FALSE, 0);
    row = 0;

    /* seed/time */
    pctrl.seed = gimp_random_seed_new(&pvals.seed, &pvals.random_seed);
    gimp_table_attach_aligned(GTK_TABLE(table), 0, row,
                              _("Random Seed:"), 0.0, 0.5,
                              pctrl.seed, 2, TRUE);
    g_signal_connect_swapped(GIMP_RANDOM_SEED_SPINBUTTON_ADJ(pctrl.seed),
                             "value-changed",
                             G_CALLBACK(execute_all), drawable);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    pctrl.freq_min = yeti_scale_entry_new_double(_("M_inimum frequency:"),
                                                 table, sgroup, row,
                                                 &pvals.freq_min,
                                                 0.0, 0.5, 0.001, 0.01, 3);
    g_signal_connect(pctrl.freq_min, "value-changed",
                     G_CALLBACK(execute_all), drawable);
    row++;

    pctrl.freq_max = yeti_scale_entry_new_double(_("Ma_ximum frequency:"),
                                                 table, sgroup, row,
                                                 &pvals.freq_max,
                                                 0.0, 0.5, 0.001, 0.01, 3);
    g_signal_connect(pctrl.freq_max, "value-changed",
                     G_CALLBACK(execute_all), drawable);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    check = gtk_check_button_new_with_mnemonic(_("Enable G_aussian ring "
                                                 "multiplier"));
    pctrl.gauss_enable = check;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), pvals.gauss_enable);
    g_signal_connect(check, "toggled", G_CALLBACK(gauss_enable_toggled), NULL);
    yeti_table_attach_to_row(table, check, row);
    row++;

    pctrl.gauss_0 = yeti_scale_entry_new_double(_("O_ffset:"),
                                                table, sgroup, row,
                                                &pvals.gauss_0,
                                                0.0, 0.25, 0.001, 0.01, 3);
    g_signal_connect(pctrl.gauss_0, "value-changed",
                     G_CALLBACK(execute_all), drawable);
    row++;

    pctrl.gauss_sigma = yeti_scale_entry_new_double(_("_Sigma:"),
                                                    table, sgroup, row,
                                                    &pvals.gauss_sigma,
                                                    0.001, 0.2, 0.001, 0.01, 3);
    g_signal_connect(pctrl.gauss_sigma, "value-changed",
                     G_CALLBACK(execute_all), drawable);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    check = gtk_check_button_new_with_mnemonic(_("Enable po_wer multiplier"));
    pctrl.power_enable = check;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), pvals.power_enable);
    g_signal_connect(check, "toggled", G_CALLBACK(power_enable_toggled), NULL);
    yeti_table_attach_to_row(table, check, row);
    row++;

    pctrl.power = yeti_scale_entry_new_double(_("Inverse _power:"),
                                              table, sgroup, row,
                                              &pvals.power,
                                              0.0, 3.0, 0.01, 0.1, 2);
    g_signal_connect(pctrl.power, "value-changed",
                     G_CALLBACK(execute_all), drawable);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    check = gtk_check_button_new_with_mnemonic(_("Enable _directional "
                                                 "multiplier"));
    pctrl.dir_enable = check;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), pvals.dir_enable);
    g_signal_connect(check, "toggled", G_CALLBACK(dir_enable_toggled), NULL);
    yeti_table_attach_to_row(table, check, row);
    row++;

    pctrl.dir_n = yeti_scale_entry_new_int(_("Num_ber of directions:"),
                                           table, sgroup, row,
                                           &pvals.dir_n, 1, 32);
    g_signal_connect(pctrl.dir_n, "value-changed",
                     G_CALLBACK(execute_all), drawable);
    row++;

    pctrl.dir_strength = yeti_scale_entry_new_double(_("Strengt_h:"),
                                                     table, sgroup, row,
                                                     &pvals.dir_strength,
                                                     0.0, 8.0, 0.01, 0.1, 2);
    g_signal_connect(pctrl.dir_strength, "value-changed",
                     G_CALLBACK(execute_all), drawable);
    row++;

    pctrl.dir_rot = yeti_scale_entry_new_double(_("Ro_tation:"),
                                                table, sgroup, row,
                                                &pvals.dir_rot,
                                                0.0, 180.0, 1.0, 5.0, 0);
    g_signal_connect(pctrl.dir_rot, "value-changed",
                     G_CALLBACK(execute_all), drawable);
    row++;

    g_object_unref(sgroup);
    gauss_enable_toggled(pctrl.gauss_enable);
    power_enable_toggled(pctrl.power_enable);
    dir_enable_toggled(pctrl.dir_enable);

    gtk_widget_show_all(dlg);
    in_param_reset = FALSE;
    execute(drawable, RECOMPUTE_ALL); /* preview image */

    ok = FALSE;
    do {
        run = gimp_dialog_run(GIMP_DIALOG(dlg));
        switch (run) {
            case RESPONSE_REVERT:
            refresh_controls(&pvals_old);
            break;

            case RESPONSE_RESET:
            refresh_controls(&pvals_defaults);
            break;

            default:
            ok = TRUE;
            break;
        }
    } while (!ok);

    gtk_widget_destroy(dlg);

    return run == GTK_RESPONSE_OK;
}

static void
gauss_enable_toggled(GtkWidget *check)
{
    pvals.gauss_enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(check));
    gimp_scale_entry_set_sensitive(pctrl.gauss_0, pvals.gauss_enable);
    gimp_scale_entry_set_sensitive(pctrl.gauss_sigma, pvals.gauss_enable);
    execute_all(drawable);
}

static void
power_enable_toggled(GtkWidget *check)
{
    pvals.power_enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(check));
    gimp_scale_entry_set_sensitive(pctrl.power, pvals.power_enable);
    execute_all(drawable);
}

static void
dir_enable_toggled(GtkWidget *check)
{
    pvals.dir_enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(check));
    gimp_scale_entry_set_sensitive(pctrl.dir_n, pvals.dir_enable);
    gimp_scale_entry_set_sensitive(pctrl.dir_strength, pvals.dir_enable);
    gimp_scale_entry_set_sensitive(pctrl.dir_rot, pvals.dir_enable);
    execute_all(drawable);
}

static void
gradient_changed(const gchar *name,
                 gint width,
                 const gdouble *grad_data,
                 gboolean dialog_closing,
                 gpointer user_data)
{
    gimp_context_set_gradient(name);
    gradient = g_realloc(gradient, width*sizeof(gdouble));
    memcpy(gradient, grad_data, width*sizeof(gdouble));
    gradient_p_width = width/4;

    execute(drawable, RECOMPUTE_COLOR);
}

static void
coloring_method_changed(GtkWidget *widget,
                        gpointer data)
{
    ColoringMethod cm;

    if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)))
        return;

    cm = GPOINTER_TO_UINT(data);
    if (pvals.cm == cm)
        return;
    gtk_widget_set_sensitive(pctrl.grad, cm == COLORING_METHOD_GRADIENT);
    pvals.cm = cm;
    execute_color(drawable);
}

/* though this is obscure, we use it to reset/revert all values _and_ refresh
   the dialog controls */
static void
refresh_controls(const PluginValues *new_pvals)
{
    in_param_reset = TRUE;

    pvals.freq_min = new_pvals->freq_min;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.freq_min),
                             pvals.freq_min);
    pvals.freq_max = new_pvals->freq_max;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.freq_max),
                             pvals.freq_max);

    pvals.gauss_enable = new_pvals->gauss_enable;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.gauss_enable),
                                 pvals.gauss_enable);
    pvals.gauss_0 = new_pvals->gauss_0;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.gauss_0),
                             pvals.gauss_0);
    pvals.gauss_sigma = new_pvals->gauss_sigma;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.gauss_sigma),
                             pvals.gauss_sigma);

    pvals.power_enable = new_pvals->power_enable;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.power_enable),
                                 pvals.power_enable);
    pvals.power = new_pvals->power;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.power),
                             pvals.power);

    pvals.dir_enable = new_pvals->dir_enable;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.dir_enable),
                                 pvals.dir_enable);
    pvals.dir_n = new_pvals->dir_n;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.dir_n),
                             pvals.dir_n);
    pvals.dir_strength = new_pvals->dir_strength;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.dir_strength),
                             pvals.dir_strength);
    pvals.dir_rot = new_pvals->dir_rot;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.dir_rot),
                             pvals.dir_rot);

    pvals.cm = new_pvals->cm;
    if (!is_rgb)
        pvals.cm = COLORING_METHOD_GRAY;
    switch (pvals.cm) {
        case COLORING_METHOD_GRAY:
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.cm_gray), TRUE);
        break;

        case COLORING_METHOD_GRADIENT:
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.cm_grad), TRUE);
        break;
    }
    gtk_widget_set_sensitive(pctrl.grad, pvals.cm == COLORING_METHOD_GRADIENT);

    pvals.seed = new_pvals->seed;
    gtk_adjustment_set_value(GIMP_RANDOM_SEED_SPINBUTTON_ADJ(pctrl.seed),
                             pvals.seed);

    /* unable to update: grad -- FIXME */
    in_param_reset = FALSE;
    execute(drawable, RECOMPUTE_ALL);
}

/****************************************************************************/

static void
execute_all(GimpDrawable *drawable)
{
    execute(drawable, RECOMPUTE_ALL);
}

static void
execute_color(GimpDrawable *drawable)
{
    execute(drawable, RECOMPUTE_COLOR);
}

static void
plugin_execute(void)
{
    gint i, j, w, h, count;
    gdouble xf, yf, f, t, phi, v, min, max, rot;

    g_rand_set_seed(rng, pvals.seed);

    w = ix2 - ix1;
    h = iy2 - iy1;
    count = 0;
    rot = pvals.dir_rot * G_PI/180.0;
    for (i = 0; i < h; i++) {
        yf = ((i <= h/2) ? i : i-h)/(gdouble)h;
        for (j = 0; j < w; j++) {
            xf = ((j <= w/2) ? j : j-w)/(gdouble)w;
            f = hypot(xf, yf);
            if (f >= pvals.freq_min && f <= pvals.freq_max) {
                v = 1.0;
                if (pvals.power_enable)
                    v *= (f == 0.0) ? 0.0 : 1.0/(pow(f, pvals.power));
                if (pvals.gauss_enable)
                    v *= exp(-(f - pvals.gauss_0)*(f - pvals.gauss_0)
                             /(2.0*pvals.gauss_sigma*pvals.gauss_sigma));
                if (pvals.dir_enable) {
                    phi = atan2(yf, xf);
                    t = (cos(pvals.dir_n*(phi - rot)) + 1.0)/2.0;
                    v *= pow(t, pvals.dir_strength);
                }
                phi = g_rand_double(rng)*2.0*G_PI;
                fft_data[i*w + j][0] = v*cos(phi);
                fft_data[i*w + j][1] = v*sin(phi);
            }
            else {
                /* Keep the number of consumed random numbers constant */
                phi = g_rand_double(rng)*2.0*G_PI;
                fft_data[i*w + j][0] = 0.0;
                fft_data[i*w + j][1] = 0.0;
            }
            count++;
            if (count % PROGRESS_STEP == 0 && !preview_mode)
                gimp_progress_update((progress + 3*count)/max_progress);
        }
    }
    if (!preview_mode) {
        progress += 3*w*h;
        gimp_progress_update(progress/max_progress);
    }
    fftw_execute(fft_plan);
    if (!preview_mode) {
        progress += 4*w*h;
        gimp_progress_update(progress/max_progress);
    }

    min = G_MAXDOUBLE;
    max = -G_MAXDOUBLE;
    for (i = 0; i < w*h; i++) {
        if (fft_data[i][0] < min)
            min = fft_data[i][0];
        if (fft_data[i][0] > max)
            max = fft_data[i][0];
    }
    if (max <= min) {
        for (i = 0; i < w*h; i++)
            fft_data[i][0] = 0.5;
    }
    else {
        for (i = 0; i < w*h; i++) {
            fft_data[i][0] = (fft_data[i][0] - min)/(max - min);
            /* Fix rounding problems */
            fft_data[i][0] = CLAMP(fft_data[i][0], 0.0, 1.0);
        }
    }
    if (!preview_mode) {
        progress += w*h;
        gimp_progress_update(progress/max_progress);
    }
}

static void
execute(GimpDrawable *drawable,
        guint recompute)
{
    gint chmax;

    if (in_param_reset)
        return;

    /* compute sizes and other stuff of this kind */
    alpha = -1000; /* to cause a segfault ;-> */
    if (preview_mode) {
        ix1 = iy1 = 0;
        ix2 = gdk_pixbuf_get_width(preview);
        iy2 = gdk_pixbuf_get_height(preview);
        bpp = gdk_pixbuf_get_n_channels(preview);
        if (!fft_data) {
            fft_data = fftw_malloc(ix2*iy2*sizeof(fftw_complex));
            /* This will be executed many times. */
            fft_plan = fftw_plan_dft_2d(ix2, iy2, fft_data, fft_data,
                                        1, FFTW_MEASURE);
        }
    }
    else {
        gimp_drawable_mask_bounds(drawable->drawable_id,
                                  &ix1, &iy1, &ix2, &iy2);
        bpp = gimp_drawable_bpp(drawable->drawable_id);
        if (has_alpha)
            alpha = bpp-1;
        if (fft_data) {
            fftw_destroy_plan(fft_plan);
            fftw_free(fft_data);
        }

        fft_data = fftw_malloc((ix2 - ix1)*(iy2 - iy1)*sizeof(fftw_complex));
        /* This will be executed once. */
        fft_plan = fftw_plan_dft_2d(iy2 - iy1, ix2 - ix1, fft_data, fft_data,
                                    1, FFTW_ESTIMATE);
    }
    max_progress = 10.0*(ix2 - ix1)*(iy2 - iy1);
    chmax = bpp;

    render_gradient();
    progress = 0;
    if ((recompute & RECOMPUTE_ALL))
        plugin_execute();
    commit(drawable, chmax);

    if (preview_mode)
        gtk_widget_queue_draw(preview_image);
    else {
        gimp_drawable_flush(drawable);
        gimp_drawable_merge_shadow(drawable->drawable_id, TRUE);
        gimp_drawable_update(drawable->drawable_id,
                             ix1, iy1, (ix2 - ix1), (iy2 - iy1));
    }
}

/* give the abstract values some visual interpretation */
static void
commit(GimpDrawable *drawable,
       gint chmax)
{
    gpointer pr;
    GimpPixelRgn dest_rgn;

    if (preview_mode) {
        dest_rgn.x = dest_rgn.y = 0;
        dest_rgn.w = dest_rgn.h = PREVIEW_SIZE;
        dest_rgn.bpp = gdk_pixbuf_get_n_channels(preview);
        dest_rgn.rowstride = gdk_pixbuf_get_rowstride(preview);
        dest_rgn.data = gdk_pixbuf_get_pixels(preview);
        commit_one_tile(&dest_rgn, chmax, gradient_p_width);
    }
    else {
        gimp_pixel_rgn_init(&dest_rgn, drawable,
                            ix1, iy1, (ix2 - ix1), (iy2 - iy1),
                            TRUE, TRUE);
        /* iterate by tiles */
        for (pr = gimp_pixel_rgns_register(1, &dest_rgn);
             pr != NULL;
             pr = gimp_pixel_rgns_process(pr)) {
            commit_one_tile(&dest_rgn, chmax, GRADIENT_SAMPLE_SIZE);
            progress += 3 * dest_rgn.w * dest_rgn.h;
            gimp_progress_update(progress/max_progress);
        }
    }
}

/* give the abstract values some visual interpretation (one tile) */
static void
commit_one_tile(GimpPixelRgn *rgn,
                gint chmax,
                gint gradwidth)
{
    static const gint check_colors[] = {
        CHANNEL_MAX_VALUE*GIMP_CHECK_DARK,
        CHANNEL_MAX_VALUE*GIMP_CHECK_LIGHT,
    };

    gint idx = 0, channel, chc;
    guint x, y;
    guchar *p;
    gdouble r = 0, f;

    for (y = rgn->y; y < rgn->y + rgn->h; y++) {
        for (x = rgn->x; x < rgn->x + rgn->w; x++) {
            /* draw the pixel */
            p = rgn->data + (y - rgn->y)*rgn->rowstride + (x - rgn->x)*rgn->bpp;
            r = fft_data[(y - iy1)*(ix2 - ix1) + (x - ix1)][0];
            if (pvals.cm == COLORING_METHOD_GRADIENT)
                idx = r*(gradwidth - 1);
            for (channel = 0; channel < chmax; channel++) {
                if (pvals.cm == COLORING_METHOD_GRADIENT) {
                    p[channel] = CHANNEL_MAX_VALUE*gradient[4*idx + channel];
                    /* add checker in preview mode */
                    if (preview_mode && has_alpha) {
                        f = gradient[4*idx + 3];
                        chc = (x - ix1)%(2*GIMP_CHECK_SIZE)/GIMP_CHECK_SIZE
                            + (y - iy1)%(2*GIMP_CHECK_SIZE)/GIMP_CHECK_SIZE;
                        p[channel] = p[channel]*f
                                     + (1.0 - f)*check_colors[chc % 2];
                    }
                }
                else {
                    if (channel == alpha)
                        p[channel] = CHANNEL_MAX_VALUE;
                    else
                        p[channel] = CHANNEL_MAX_VALUE*r;
                }
            }
        }
    }
}

static void
render_gradient(void)
{
    gint dummy;

    if (pvals.cm != COLORING_METHOD_GRADIENT)
        return;

    if (preview_mode) {
        /* don't regenerate what we already have */
        if (gradient == NULL) {
            gradient_p_width = PREVIEW_SIZE;
            gimp_gradient_get_uniform_samples(gimp_context_get_gradient(),
                                              gradient_p_width,
                                              FALSE,
                                              &dummy, &gradient);
        }
    }
    else {
        /* final rendering, done only once at the end */
        g_free(gradient);
        gimp_gradient_get_uniform_samples(gimp_context_get_gradient(),
                                          GRADIENT_SAMPLE_SIZE,
                                          FALSE,
                                          &dummy, &gradient);
    }
}

